#include <tcl.h>
#include "Str.h"
#include "Term.h"

static int
usage(Tcl_Interp *interp)
{
    Tcl_AppendResult(interp, "incorrect usage, \"term help\" for more information", NULL);
    return TCL_ERROR;
}

static void
help(Tcl_Interp *interp)
{
    Tcl_AppendResult(interp, "usage:\n"\
     "   term <cmd>\n\n"\
     "   cmds: \n"\
     "      init [raw]   initilizes terminal\n"\
     "      exit         revert terminal to original state\n"\
     "      enter        use upon returning from suspension\n"\
     "      can_scroll   returns 1 if terminal can scroll, 0 if not\n"\
     "      flush        flush buffer\n"\
     "      corner       flush buffer and move cursor to lower right\n"\
     "      beep [vis]   beep, visually if allowed\n"\
     "      screensize   reconfig after sigwinch\n"\
     "      clear        clear screen\n"\
     "      clreol       clear to end of line\n"\
     "      move <r> <c> move cursor\n"\
     "      puts <str>   put string\n"\
     "      gets [sz]    get string, with optional scrollsize\n"\
     "      getchar      get one character\n"\
     "      rows         get number of rows\n"\
     "      cols         get number of columns\n"\
     "      curr_row     get current cursor row\n"\
     "      curr_col     get current cursor column\n"\
     "      scroll <t> <b> <#> scroll by dir # rows, can be negative.\n"
     "                         t is top and b is the bottom of the\n"
     "                         scroll region.\n"\
     "      attr <a> <fg> <bg> set attributes and colors. 0 to skip field."
     "                         attributes are: bold, underline, inverse, all\n"
     "                         append 'off' to attributes to turn them off.\n"
     "                         colors are: red, green, blue, black, cyan,\n"
     "                                     magenta, yellow, white, default.\n",
    NULL);
}

static int
terminput(ClientData area, Tcl_Interp *interp, int argc, char *argv[])
{
    if (argc < 2)
	return usage(interp);

    if (argc == 2 && !strcasecmp(argv[1], "free")) {
	Tcl_DeleteCommand(interp, argv[0]);
	Term_InputFree(area);
	return TCL_OK;

    } else if (argc == 3 && !strcasecmp(argv[1], "addhistory")) {
	Term_InputAddHistory(area, argv[2]);
	return TCL_OK;

    } else if (argc == 4 && !strcasecmp(argv[1], "addmacro")) {
	Term_InputAddMacro(area, argv[2], argv[3]);
	return TCL_OK;

    } else if (argc == 2 && !strcasecmp(argv[1], "movecursor")) {
	Term_InputMoveCursor(area);
	return TCL_OK;

    } else if (argc == 5 && !strcasecmp(argv[1], "reposition")) {
	int r = atoi(argv[2]);
	int c = atoi(argv[3]);
	int w = atoi(argv[4]);
	Term_InputReposition(area, r, c, w);
	return TCL_OK;

    } else if (argc == 2 && !strcasecmp(argv[1], "redraw")) {
	Term_InputRedraw(area);
	return TCL_OK;

    } else if (argc == 2 && !strcasecmp(argv[1], "clear")) {
	Term_InputClear(area);
	return TCL_OK;

    } else if (argc == 3 && !strcasecmp(argv[1], "insert")) {
	Term_InputInsert(area, argv[2]);
	return TCL_OK;

    } else if (argc == 4 && !strcasecmp(argv[1], "process")) {
	char buf[20];
	char *p;
	int i = Term_InputProcess(area, argv[2][0], &p);
	if (i == 1 || i == 2)
	    Tcl_SetVar(interp, argv[3], p, 0);
	sprintf(buf, "%d", i);

	Tcl_AppendResult(interp, buf, NULL);
	return TCL_OK;

    } else {
	return TCL_ERROR;
    }
} 

static int
term(ClientData notused, Tcl_Interp *interp, int argc, char *argv[])
{
    if (argc < 2)
	return usage(interp);

    if (argc == 2 && !strcasecmp(argv[1], "help")) {
	help(interp);
	return TCL_OK;

    } else if (argc <= 3 && !strcasecmp(argv[1], "init")) {
	Term_Init(argc == 3);
	return TCL_OK;

    } else if (argc == 2 && !strcasecmp(argv[1], "exit")) {
	Term_Exit(TERM_CLS_RESTORE);
	return TCL_OK;

    } else if (argc == 2 && !strcasecmp(argv[1], "enter")) {
	Term_Enter();
	return TCL_OK;

    } else if (argc == 2 && !strcasecmp(argv[1], "flush")) {
	Term_Flush();
	return TCL_OK;

    } else if (argc == 2 && !strcasecmp(argv[1], "corner")) {
	Term_Corner();
	return TCL_OK;

    } else if (argc <= 3 && !strcasecmp(argv[1], "beep")) {
	Term_Beep(argc == 3);
	return TCL_OK;

    } else if (!strcasecmp(argv[1], "screensize")) {
	Term_ScreenSize();
	return TCL_OK;

    } else if (argc == 2 && !strcasecmp(argv[1], "clear")) {
	Term_Clear();
	return TCL_OK;

    } else if (argc == 2 && !strcasecmp(argv[1], "clreol")) {
	Term_ClearToEOL();
	return TCL_OK;

    } else if (argc == 4 && !strcasecmp(argv[1], "move")) {
	Term_Move(atoi(argv[2]), atoi(argv[3]));
	return TCL_OK;

    } else if (argc == 3 && !strcasecmp(argv[1], "puts")) {
	Term_Puts(argv[2]);
	return TCL_OK;

    } else if (argc == 2 && !strcasecmp(argv[1], "can_scroll")) {
	int i, ret = TCL_ERROR;
	if (Term_Can_Scroll)
	    Tcl_AppendResult(interp, "1", NULL);
	else
	    Tcl_AppendResult(interp, "0", NULL);

	return TCL_OK;

    } else if (argc <= 3 && !strcasecmp(argv[1], "gets")) {
	int i, ret = TCL_ERROR;
	Str *s = NULL;
	if (argc == 2) {
	    s = StrNew();
	    if (Term_GetInput(s, Term_Cols() - 1 - Term_CurrCol())) {
		Tcl_AppendResult(interp, StrStr(s), NULL);
		ret = TCL_OK;
	    }
	} else if ((i = atoi(argv[2]))) {
	    s = StrNew();
	    if (Term_GetInput(s, i) >= 0) {
		Tcl_AppendResult(interp, StrStr(s), NULL);
		ret = TCL_OK;
	    }
	}
	StrFree(s);
	return ret;

    } else if (argc == 2 && !strcasecmp(argv[1], "getchar")) {
	int ch = Term_GetChar();
	if (ch > 0 && ch <= 255) {
	    char buf[2];
	    buf[1] = 0;
	    buf[0] = ch;
	    Tcl_AppendResult(interp, buf, NULL);
	}
	return TCL_OK;

    } else if (argc <= 2 && !strcasecmp(argv[1], "rows")) {
	char buf[10];
	snprintf(buf, sizeof(buf), "%d", Term_Rows());
	Tcl_AppendResult(interp, buf, NULL);
	return TCL_OK;

    } else if (argc <= 2 && !strcasecmp(argv[1], "cols")) {
	char buf[10];
	snprintf(buf, sizeof(buf), "%d", Term_Cols());
	Tcl_AppendResult(interp, buf, NULL);
	return TCL_OK;

    } else if (argc <= 2 && !strcasecmp(argv[1], "curr_row")) {
	char buf[10];
	snprintf(buf, sizeof(buf), "%d", Term_CurrRow());
	Tcl_AppendResult(interp, buf, NULL);
	return TCL_OK;

    } else if (argc <= 2 && !strcasecmp(argv[1], "curr_col")) {
	char buf[10];
	snprintf(buf, sizeof(buf), "%d", Term_CurrCol());
	Tcl_AppendResult(interp, buf, NULL);
	return TCL_OK;

    } else if (argc == 5 && !strcasecmp(argv[1], "scroll")) {
	int j, i = atoi(argv[4]);
	if (i == 0)
	    return TCL_OK;
	else if (i < 0)
	    j = -i;
	else
	    j = i;

	if (Term_Scroll(atoi(argv[2]), atoi(argv[3]), i) != j)
	    return TCL_ERROR;
	return TCL_OK;

    } else if (argc == 5 && !strcasecmp(argv[1], "attr")) {
	int attr = -1, fg = -1, bg = -1;
	int attroff = -1;
	if (!strcasecmp(argv[2], "0"))              attr = 0;
	else if (!strcasecmp(argv[2], "bold"))      attr = ATTR_BOLD;
	else if (!strcasecmp(argv[2], "underline")) attr = ATTR_UNDERLINE;
	else if (!strcasecmp(argv[2], "inverse"))   attr = ATTR_INVERSE;
	else if (!strcasecmp(argv[2], "all"))       attr = ATTR_INVERSE|ATTR_UNDERLINE|ATTR_BOLD;
	else {
	    attr = 0;
	    if (!strcasecmp(argv[2], "boldoff"))      	   attroff = ATTR_BOLD;
	    else if (!strcasecmp(argv[2], "underlineoff")) attroff = ATTR_UNDERLINE;
	    else if (!strcasecmp(argv[2], "inverseoff"))   attroff = ATTR_INVERSE;
	    else if (!strcasecmp(argv[2], "alloff"))       attroff = ATTR_INVERSE|ATTR_UNDERLINE|ATTR_BOLD;
	    else return TCL_ERROR;
	}

	if (!strcasecmp(argv[3], "0"))            fg = 0;
	else if (!strcasecmp(argv[3], "red"))     fg = ATTR_COLOR_RED;
	else if (!strcasecmp(argv[3], "blue"))    fg = ATTR_COLOR_BLUE;
	else if (!strcasecmp(argv[3], "green"))   fg = ATTR_COLOR_GREEN;
	else if (!strcasecmp(argv[3], "magenta")) fg = ATTR_COLOR_MAGENTA;
	else if (!strcasecmp(argv[3], "cyan"))    fg = ATTR_COLOR_CYAN;
	else if (!strcasecmp(argv[3], "black"))   fg = ATTR_COLOR_BLACK;
	else if (!strcasecmp(argv[3], "white"))   fg = ATTR_COLOR_WHITE;
	else if (!strcasecmp(argv[3], "yellow"))  fg = ATTR_COLOR_YELLOW;
	else if (!strcasecmp(argv[3], "default")) fg = ATTR_COLOR_DEFAULT;
	else return TCL_ERROR;

	if (!strcasecmp(argv[4], "0"))            bg = 0;
	else if (!strcasecmp(argv[4], "red"))     bg = ATTR_COLOR_RED;
	else if (!strcasecmp(argv[4], "blue"))    bg = ATTR_COLOR_BLUE;
	else if (!strcasecmp(argv[4], "green"))   bg = ATTR_COLOR_GREEN;
	else if (!strcasecmp(argv[4], "magenta")) bg = ATTR_COLOR_MAGENTA;
	else if (!strcasecmp(argv[4], "cyan"))    bg = ATTR_COLOR_CYAN;
	else if (!strcasecmp(argv[4], "black"))   bg = ATTR_COLOR_BLACK;
	else if (!strcasecmp(argv[4], "white"))   bg = ATTR_COLOR_WHITE;
	else if (!strcasecmp(argv[4], "yellow"))  bg = ATTR_COLOR_YELLOW;
	else if (!strcasecmp(argv[4], "default")) bg = ATTR_COLOR_DEFAULT;
	else return TCL_ERROR;

	if (attroff != -1)
	    Term_Attr_Off(attroff);

	Term_Attr(attr, fg, bg);

	return TCL_OK;

    } else if (argc == 9 && !strcasecmp(argv[1], "newinput")) {
	static int cmdid = 0;
	char buf[64];
	int r = atoi(argv[2]);
	int c = atoi(argv[3]);
	int w = atoi(argv[4]);
	int h = atoi(argv[5]);
	int attr, fg, bg;
	void *p;

	if (!strcasecmp(argv[6], "none"))	    attr = 0;
	else if (!strcasecmp(argv[6], "bold"))      attr = ATTR_BOLD;
	else if (!strcasecmp(argv[6], "underline")) attr = ATTR_UNDERLINE;
	else if (!strcasecmp(argv[6], "inverse"))   attr = ATTR_INVERSE;
	else if (!strcasecmp(argv[6], "all"))       attr = ATTR_INVERSE|ATTR_UNDERLINE|ATTR_BOLD;
	else return TCL_ERROR;

	if (!strcasecmp(argv[7], "red"))	  fg = ATTR_COLOR_RED;
	else if (!strcasecmp(argv[7], "blue"))    fg = ATTR_COLOR_BLUE;
	else if (!strcasecmp(argv[7], "green"))   fg = ATTR_COLOR_GREEN;
	else if (!strcasecmp(argv[7], "magenta")) fg = ATTR_COLOR_MAGENTA;
	else if (!strcasecmp(argv[7], "cyan"))    fg = ATTR_COLOR_CYAN;
	else if (!strcasecmp(argv[7], "black"))   fg = ATTR_COLOR_BLACK;
	else if (!strcasecmp(argv[7], "white"))   fg = ATTR_COLOR_WHITE;
	else if (!strcasecmp(argv[7], "yellow"))  fg = ATTR_COLOR_YELLOW;
	else if (!strcasecmp(argv[7], "default")) fg = ATTR_COLOR_DEFAULT;
	else return TCL_ERROR;

	if (!strcasecmp(argv[8], "red"))	  bg = ATTR_COLOR_RED;
	else if (!strcasecmp(argv[8], "blue"))    bg = ATTR_COLOR_BLUE;
	else if (!strcasecmp(argv[8], "green"))   bg = ATTR_COLOR_GREEN;
	else if (!strcasecmp(argv[8], "magenta")) bg = ATTR_COLOR_MAGENTA;
	else if (!strcasecmp(argv[8], "cyan"))    bg = ATTR_COLOR_CYAN;
	else if (!strcasecmp(argv[8], "black"))   bg = ATTR_COLOR_BLACK;
	else if (!strcasecmp(argv[8], "white"))   bg = ATTR_COLOR_WHITE;
	else if (!strcasecmp(argv[8], "yellow"))  bg = ATTR_COLOR_YELLOW;
	else if (!strcasecmp(argv[8], "default")) bg = ATTR_COLOR_DEFAULT;
	else return TCL_ERROR;

	p = Term_InputInit(r, c, w, h, attr, fg, bg);

	sprintf(buf, "term_input%d", cmdid++);
	Tcl_CreateCommand(interp, buf, terminput, p, 0);
	Tcl_AppendResult(interp, buf, NULL);
	return TCL_OK;

    } else {
	return TCL_ERROR;
    }
}

int
Tclterm_Init(Tcl_Interp *interp)
{
    int ret;
    if ((ret=Tcl_PkgProvide(interp, "Term", "1.0")) != TCL_OK)
	return ret;

    Tcl_CreateCommand (interp, "term", term, 0, 0);
    return TCL_OK;
}
